#!/usr/bin/env python3

import sys
import optparse
import re
import markdown
import json

# https://pannellum.org/documentation/overview/

popt = optparse.OptionParser()

popt.add_option('-n', '--dry-run', '--dryrun',
                action='store_true', dest='dryrun')
popt.add_option('-v', '--verbose',
                action='store_true', dest='verbose')

(opts, args) = popt.parse_args()

class TextChapter:
    def __init__(self, index, html, pos):
        self.index = index
        self.html = html
        self.pos = pos

class Book:
    booklist = []
    bookmap = {}
    
    def __init__(self, key):
        self.key = key
        self.spots = []
        self.snippets = []
        self.html = None
        self.title = None
        self.author = None
        self.item = None
        self.itemhtml = '<p>???</p>'
        if key in Book.bookmap:
            raise Exception('duplicate book: %s' % (key,))
        Book.booklist.append(self)
        Book.bookmap[key] = self

    def __repr__(self):
        return '<Book %s>' % (self.key,)

class Snippet:
    snippetlist = []
    snippetmap = {}  # indexed by (bookkey, itemkey)

    def __init__(self, bookkey, itemkey):
        self.bookkey = bookkey
        self.itemkey = itemkey
        self.html = None
        tup = (bookkey, itemkey)
        if tup in Snippet.snippetmap:
            raise Exception('Duplicate snippet: %s' % (tup,))
        Snippet.snippetlist.append(self)
        Snippet.snippetmap[tup] = self
        Book.bookmap[bookkey].snippets.append(self)
        
    def __repr__(self):
        return '<Snippet %s:%s>' % (self.bookkey, self.itemkey,)
        
class Node:
    nodelist = []
    nodemap = {}
    
    def __init__(self, key, northoffset=0, startyaw=None):
        self.key = key
        self.startyaw = startyaw
        self.filename = 'pic/library-%s.jpeg' % (key.replace('_', '-'),)
        self.northoffset = northoffset
        self.navspots = []
        self.textspots = []
        self.bookspots = []
        Node.nodelist.append(self)
        Node.nodemap[key] = self

    def __repr__(self):
        return '<Node %s>' % (self.key,)

class NavSpot:
    def __init__(self, key, yaw, pitch=0):
        self.key = key
        self.yaw = yaw
        self.pitch = pitch
        val = 'LowSpot' if key.endswith('low') else 'HighSpot'
        self.cssclass = ' '.join([ 'pnlm-hotspot pnlm-sprite pnlm-scene', val ])

    def write(self, fl, comma=''):
        fl.write('      {\n')
        fl.write('        pitch: %d, yaw: %d,\n' % (self.pitch, self.yaw,))
        fl.write('        type: \'scene\',\n')
        fl.write('        sceneId: %r,\n' % (self.key,))
        fl.write('        targetYaw: \'sameAzimuth\',\n')
        fl.write('        targetPitch: \'same\',\n')
        fl.write('        cssClass: %r,\n' % (self.cssclass,))
        fl.write('        scale: true\n')
        fl.write('      }%s\n' % (comma,))

class BookSpot:
    def __init__(self, key, nodekey, yaw, pitch, size=None, skewpitch=None, skewyaw=None, skewoffsetx=None):
        self.key = key
        self.nodekey = nodekey
        self.fullkey = 'spot_%s_%s' % (nodekey, key,)
        self.yaw = yaw
        self.pitch = pitch
        self.size = size
        self.skewpitch = skewpitch
        self.skewyaw = skewyaw
        self.skewoffsetx = skewoffsetx
        val = 'book_' + key
        self.cssclass = ' '.join([ 'pnlm-hotspot pnlm-sprite pnlm-info', 'BookSpot', val, self.fullkey ])

    def write(self, fl, comma=''):
        fl.write('      {\n')
        fl.write('        pitch: %.1f, yaw: %.1f,\n' % (self.pitch, self.yaw,))
        fl.write('        type: \'info\',\n')
        fl.write('        cssClass: %r,\n' % (self.cssclass,))
        fl.write('        clickHandlerFunc: clickfunc,\n')
        fl.write('        clickHandlerArgs: { type:\'book\', key:%r },\n' % (self.key,))
        if self.skewpitch is not None:
            fl.write('        skewpitch: %.1f,\n' % (self.skewpitch,))
        if self.skewyaw is not None:
            fl.write('        skewyaw: %.1f,\n' % (self.skewyaw,))
        if self.skewoffsetx is not None:
            fl.write('        skewoffsetx: %.1f,\n' % (self.skewoffsetx,))
        fl.write('        scale: true\n')
        fl.write('      }%s\n' % (comma,))

    def __repr__(self):
        return '<Node %s (%s)>' % (self.key, self.nodekey)

class TextSpot:
    def __init__(self, nodekey, index, yaw, pitch, scale=1.0, skewyaw=0.0):
        self.nodekey = nodekey
        self.index = index
        self.scale = scale
        self.yaw = yaw
        self.pitch = pitch
        self.skewyaw = skewyaw
        self.fullkey = 'spot_%s__text_%s' % (nodekey, index)
        self.cssclass = ' '.join([ 'pnlm-hotspot pnlm-sprite pnlm-info', 'TextSpot', 'TextSpot_%d' % (index,), self.fullkey ])

    def write(self, fl, comma=''):
        fl.write('      {\n')
        fl.write('        pitch: %.1f, yaw: %.1f,\n' % (self.pitch, self.yaw,))
        fl.write('        type: \'info\',\n')
        fl.write('        cssClass: %r,\n' % (self.cssclass,))
        fl.write('        clickHandlerFunc: clickfunc,\n')
        fl.write('        clickHandlerArgs: { type:\'text\', index:%r },\n' % (self.index,))
        fl.write('        skewpitch: %.1f,\n' % (0.0,))
        fl.write('        skewyaw: %.1f,\n' % (self.skewyaw,))
        fl.write('        scale: true\n')
        fl.write('      }%s\n' % (comma,))

def commify(ls, comma=','):
    pos = 0
    count = len(ls)
    while pos < count:
        yield (ls[pos], comma if (pos+1<count) else '')
        pos += 1

converter = markdown.Markdown(extensions = ['meta'])

def readmap(fl):
    pat_line = re.compile('^(\s*)([^: ]*):(.*)$')

    curnode = None
    
    for ln in fl.readlines():
        ln = ln.rstrip()
        if not ln or ln.startswith('#'):
            continue
        match = pat_line.match(ln)
        if not match:
            raise Exception('invalid map line: %s' % (ln,))
        indent = bool(match.group(1))
        key = match.group(2).strip()
        if key.startswith('#'):
            continue
        args = match.group(3).split(',')
        args = [ int(val.strip()) for val in args ]

        if not indent:
            northoffset = args[0]
            startyaw = args[1] if len(args) >= 2 else None
            curnode = Node(key, northoffset, startyaw=startyaw)
            continue

        if key.startswith('_TEXT_'):
            val = int(key[6:])
            yaw = args[0]
            pitch = args[1]
            scale = 1.0
            if len(args) >= 3:
                scale = args[2] * 0.1
            skewyaw = 0.0
            if len(args) >= 4:
                skewyaw = float(args[3])
            spot = TextSpot(curnode.key, val, yaw, pitch, scale=scale, skewyaw=skewyaw)
            #if mainchapters[val].pos != curnode.key:
            #    raise Exception('text spot in wrong position (spot %d is %s, chapter is at %s)' % (val, curnode.key, mainchapters[val].pos,))
            curnode.textspots.append(spot)
        else:
            yaw = args[0]
            pitch = args[1] if len(args) >= 2 else 0
            spot = NavSpot(key, yaw, pitch)
            curnode.navspots.append(spot)

def readbooks(fl):
    pat_snippet = re.compile('^[.][.][.]([^:]*)\\s*:\\s*$')
    pat_divider = re.compile('^-----[-]*$')
    ls = []
    curkey = None
    curusekey = None
    
    for ln in fl.readlines():
        ln = ln.rstrip()

        if pat_divider.match(ln):
            continue
        
        match = pat_snippet.match(ln)
        
        if ln.startswith('##'):
            parseblock(ls, curkey, curusekey)
            curkey = ln[2:].strip()
            curusekey = None
            ls = []

        elif match:
            parseblock(ls, curkey, curusekey)
            # curkey stays the same
            curusekey = match.group(1).strip()
            ls = []
            
        else:
            ls.append(ln)

    parseblock(ls, curkey, curusekey)

def parseblock(ls, key, usekey=None):
    if not key:
        return
    
    text = '\n'.join(ls)
    if not text.strip():
        return

    if not usekey:
        book = Book(key)
        book.html = converter.convert(text)
        metadata = dict(converter.Meta)
        book.title = ''.join(metadata.get('title', '###'))
        book.author = ''.join(metadata.get('author', '###'))
        atls = metadata.get('at')
        if atls:
            for val in atls:
                posls = val.split(',')
                if len(posls) < 3:
                    raise Exception('Not enough values in pos: %s' % (val,))
                nodekey, pitch, yaw = posls[0:3]
                size = None
                if len(posls) >= 4:
                    size = posls[3].split('x')
                    size = [ int(val) for val in size ]
                pitch = float(pitch.strip())
                yaw = float(yaw.strip())
                hsargs = {}
                for arg in posls[4:]:
                    arg = arg.strip()
                    argkey, _, argval = arg.partition('=')
                    if argkey == 'skewyaw':
                        hsargs['skewyaw'] = float(argval)
                    elif argkey == 'skewpitch':
                        hsargs['skewpitch'] = float(argval)
                    elif argkey == 'skewoffsetx':
                        hsargs['skewoffsetx'] = float(argval)
                    elif argkey == 'skewoffsety':
                        hsargs['skewoffsety'] = float(argval)
                    else:
                        raise Exception('unrecognized hotspot arg: %s' % (arg,))
                spot = BookSpot(key, nodekey, yaw, pitch, size=size, **hsargs)
                book.spots.append(spot)
        valls = metadata.get('has')
        if valls:
            if len(valls) != 1:
                raise Exception('has must be "key, name"')
            itemkey, itemname = valls[0].split(',')
            book.item = itemkey.strip()
            book.itemname = itemname.strip()
        valls = metadata.get('hasdesc')
        if valls:
            book.itemhtml = converter.convert('\n'.join(valls))
    else:
        snip = Snippet(key, usekey)
        snip.html = converter.convert(text)
        snip.islong = (len(snip.html) >= 300)
        
mainchapters = []

def readmaintext(fl):
    ls = []
    curkey = None
    
    for ln in fl.readlines():
        ln = ln.rstrip()
        if not ln.startswith('##'):
            ls.append(ln)
            continue

        newcurkey = ln[2:].strip()

        text = '\n'.join(ls)
        if curkey:
            if int(curkey) != len(mainchapters):
                raise Exception('Chapter index does not match')
            val = converter.convert(text)
            pos = converter.Meta.get('at')[0]
            mainchapters.append(TextChapter(int(curkey), val, pos=pos))

        ls = []
        curkey = newcurkey

def writespots(fl):
    fl.write('\'use strict\';\n\n')
    fl.write('function clickfunc(ev, args) { evhan_infoclick(ev, args); }\n\n')
    fl.write('var scenelist = {\n')
    for node, comma in commify(Node.nodelist):
        fl.write('  %s: {\n' % (node.key,))
        fl.write('    panorama: %r,\n' % (node.filename,))
        if node.startyaw is not None:
            fl.write('    yaw: %d,\n' % (node.startyaw,))
        fl.write('    northOffset: %d,\n' % (node.northoffset,))
        fl.write('    hotSpots: [\n')
        spots = node.navspots + node.textspots + node.bookspots
        for spot, comma2 in commify(spots):
            spot.write(fl, comma2)
        fl.write('    ]\n')
        fl.write('  }%s\n' % (comma,))
    fl.write('};\n')

def writedata(fl):
    fl.write('\'use strict\';\n\n')
    fl.write('var chaptertext = [\n')
    for chap, comma in commify(mainchapters):
        fl.write('  %s%s\n' % (json.dumps(chap.html), comma,))
    fl.write('];\n')
    
    fl.write('var bookmap = {\n')
    for book, comma in commify(Book.booklist):
        fl.write('  %s: {\n' % (book.key,))
        fl.write('    key: %r,\n' % (book.key,))
        fl.write('    title: %s,\n' % (json.dumps(book.title),))
        fl.write('    author: %s,\n' % (json.dumps(book.author),))
        if book.key == 'the_road_to_san_giovanni':
            fl.write('    containslens: true,\n')
        fl.write('    html: %s,\n' % (json.dumps(book.html),))
        if book.item:
            fl.write('    item: %r,\n' % (book.item,))
        fl.write('    snippetcount: %d,\n' % (len(book.snippets),))
        fl.write('    snippets: {\n')
        for snip, comma2 in commify(book.snippets):
            fl.write('      %r: {\n' % (snip.itemkey,))
            fl.write('        html: %s\n' % (json.dumps(snip.html),))
            fl.write('      }%s\n' % (comma2,))
        fl.write('    }\n')
        fl.write('  }%s\n' % (comma,))
    fl.write('};\n')

    ls = [ book for book in Book.booklist if book.item ]
    fl.write('var itemmap = {\n')
    for book, comma in commify(ls):
        fl.write('  %s: {\n' % (book.item,))
        fl.write('    key: %r, book: %r,\n' % (book.item, book.key,))
        fl.write('    name: %s,\n' % (json.dumps(book.itemname),))
        fl.write('    html: %s\n' % (json.dumps(book.itemhtml),))
        fl.write('  }%s\n' % (comma,))
    fl.write('};\n')
    
fl = open('maintext')
readmaintext(fl)
fl.close()

fl = open('map')
readmap(fl)
fl.close()

fl = open('allbooks')
readbooks(fl)
fl.close()

sizestanzas = []

itemset = set([ book.item for book in Book.booklist if book.item ])
for snip in Snippet.snippetlist:
    if snip.bookkey not in Book.bookmap:
        raise Exception('Book not found: %s' % (snip,))
    if snip.itemkey not in itemset:
        raise Exception('Item not found: %s' % (snip,))
    
for book in Book.booklist:
    for spot in book.spots:
        node = Node.nodemap.get(spot.nodekey)
        if not node:
            raise Exception('Node not found: %s (for %s)' % (spot.nodekey, book.key,))
        node.bookspots.append(spot)
        if spot.size:
            sizestanzas.append( (spot.fullkey, spot.size[0], spot.size[1]) )

ls = [ book for book in Book.booklist if book.item ]
lslong = [ snip for snip in Snippet.snippetlist if snip.islong ]
print('%d books, %d items, %d snippets (%d long)' % (len(Book.booklist), len(ls), len(Snippet.snippetlist), len(lslong),) )

if opts.verbose:
    map = {}
    print('book snippets:')
    for book in Book.booklist:
        for snip in book.snippets:
            ls = map.get(snip.itemkey)
            if not ls:
                ls = []
                map[snip.itemkey] = ls
            ls.append(snip)
        sval = '*' if book.item else ''
        lslong = [ snip for snip in book.snippets if snip.islong ]
        val = '%d (%dL)' % (len(book.snippets), len(lslong),)
        print(' ', val, sval+book.key)
    print('item snippets:')
    for book in Book.booklist:
        if book.item:
            ls = map.get(book.item, [])
            lslong = [ snip for snip in ls if snip.islong ]
            val = '%d (%dL)' % (len(ls), len(lslong),)
            print(' ', val, book.item)

if False:
    itemlist = [ book.item for book in Book.booklist if book.item ]
    invmap = {}
    for ix, item in enumerate(itemlist):
        invmap[item] = ix
    for book in Book.booklist:
        ls = [ '.' ] * len(itemlist)
        if book.item is not None:
            ix = invmap[book.item]
            ls[ix] = '\\'
        for snip in book.snippets:
            ix = invmap[snip.itemkey]
            ls[ix] = '*'
        print(''.join(ls))
            
if not opts.dryrun:
    fl = open('spots.js', 'w')
    writespots(fl)
    fl.close()
    
    fl = open('spots.css', 'w')
    for (key, sizex, sizey) in sizestanzas:
        # Hotspots originally sized for 960x600, almost, not exactly
        fl.write('.BookSpot.%s { width:%.03f%%; height:%.03f%%; }\n' % (key, sizex * 0.10266, sizey * 0.16447))
    for node in node.nodelist:
        for spot in node.textspots:
            fl.write('.TextSpot.%s { width:%.03f%%; height:%.03f%%; }\n' % (spot.fullkey, 2.5*spot.scale, 5.3*spot.scale,))
    fl.close()
    
    fl = open('data.js', 'w')
    writedata(fl)
    fl.close()
